/*********************************************************************/
/* file: color.c - functions related to the color command            */
/*                             HUT TINTIN                            */
/*                        coded by Ossi Tulonen                      */
/*********************************************************************/
/*
 * Copyright (c) 1992 1993 1994 1995 Ossi tulonen
 * All rights reserved.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */
 
#include "structs.h"

#include <string.h>
#include <assert.h>
#include <ctype.h>

#include "color.h"
#include "llist.h"
#include "parse.h"
#include "screen.h"
#include "session.h"
#include "utility.h"
#include "utils.h"

unsigned short default_color = DEFCOL_MODE;

extern struct master_table_data master_table;

const char *color_names[16] = { // DO NOT CHANGE !!! or if you change make sure
  "black",                      // X-Window knows that color name.
  "red",
  "green",
  "yellow",
  "blue",
  "magenta",
  "cyan",
  "white",  /* 0 - 7 */
  "white",  /* default foreground */
  "black",  /* default background */
  "grey90", /* default xcol1      */
  "grey80", /* default xcol2      */
  "grey70", /* default xcol3      */
  "grey60", /* default xcol4      */
  "grey50", /* default xcol5      */
  "grey40"  /* default xcol6      */  /* 8 - 15 */
  };

int color_nums[16] = {
  30, 31, 32, 33, 34, 35, 36, 37,
  38, 39, 50, 51, 52, 53, 54, 55
  };

const unsigned char *colors_str[] = {
  (const uchar *)"reset",          (const uchar *)"bold",         /*  0 -  1 */
  (const uchar *)"faint",          (const uchar *)"italicised",
  (const uchar *)"underlined",     (const uchar *)"blink",
  (const uchar *)"",               (const uchar *)"negative",
  (const uchar *)"",               (const uchar *)"",             /*  8 -  9 */
  (const uchar *)"",               (const uchar *)"",
  (const uchar *)"",               (const uchar *)"",
  (const uchar *)"",               (const uchar *)"",
  (const uchar *)"",               (const uchar *)"",
  (const uchar *)"",               (const uchar *)"",             /* 18 - 19 */
  (const uchar *)"",               (const uchar *)"",
  (const uchar *)"bold_off",       (const uchar *)"italicised_off",
  (const uchar *)"underlined_off", (const uchar *)"blink_off",
  (const uchar *)"",               (const uchar *)"negative_off",
  (const uchar *)"",               (const uchar *)"",             /* 28 - 29 */
  (const uchar *)"black",          (const uchar *)"red",
  (const uchar *)"green",          (const uchar *)"yellow",
  (const uchar *)"blue",           (const uchar *)"violet",
  (const uchar *)"cyan",           (const uchar *)"white",
  (const uchar *)"",               (const uchar *)"",             /* 38 - 39 */
  (const uchar *)"BLACK",          (const uchar *)"RED",
  (const uchar *)"GREEN",          (const uchar *)"YELLOW",
  (const uchar *)"BLUE",           (const uchar *)"VIOLET",
  (const uchar *)"CYAN",           (const uchar *)"WHITE",
  (const uchar *)"",               (const uchar *)"",             /* 40 - 49 */
  (const uchar *)"xcol1",          (const uchar *)"xcol2",
  (const uchar *)"xcol3",          (const uchar *)"xcol4",
  (const uchar *)"xcol5",          (const uchar *)"xcol6",
  (const uchar *)"",               (const uchar *)"",
  (const uchar *)"",               (const uchar *)"",             /* 50 - 59 */
  (const uchar *)"XCOL1",          (const uchar *)"XCOL2",
  (const uchar *)"XCOL3",          (const uchar *)"XCOL4",
  (const uchar *)"XCOL5",          (const uchar *)"XCOL6",
  (const uchar *)"",               (const uchar *)"",
  (const uchar *)"",               (const uchar *)"",             /* 60 - 69 */
  (const uchar *)"save",           (const uchar *)"load"          /* 70 - 71 */
  };

#define COLOR_NUM 72

/* ************************************************************
   the #color command
************************************************************ */

struct session *comm_color(const unsigned char *arg, session *ses)
{
  unsigned char left[BUFFER_SIZE], right[BUFFER_SIZE], buf[BUFFER_SIZE];
  unsigned short tmp_colormode;
  int i, j, pos = 0, found;
  listnode **mycolors, **myvars, *ln;
  
  if(strcmp((const char *)arg, "help") == 0) {
    ses->print((const uchar *)"#COLORS:\n\r");
    for(i = 0, j = 0; i < COLOR_NUM; i++) {
      if(*colors_str[i]) {
        if(!j)
          pos = str_cat(buf, 0, sizeof(buf), (const uchar *)"        ");
	
	pos = str_cat(buf, pos, sizeof(buf), (const uchar *)"  ");
        sprintf((char *)left, "\033[%dm", i);
	pos = str_cat(buf, pos, sizeof(buf), left);
	tmp_colormode = ses->get_output()->get_colormode();
	color_to_bits(left, tmp_colormode, ses->get_output());
	sprintf((char *)left, "%10s", colors_str[i]);
        pos = str_cat(buf, pos, sizeof(buf), left);
	bits_to_color(tmp_colormode, ses->get_output()->get_colormode(), left);
	pos = str_cat(buf, pos, sizeof(buf), left);
		      
	if(++j == 5) {
	  pos = str_cat(buf, pos, sizeof(buf), (const uchar *)"\n\r");
          ses->print(buf);
          j = 0;
        }
      }
    }
    if(j) {
      pos = str_cat(buf, pos, sizeof(buf), (const uchar *)"\n\r");
      ses->print(buf);
    }
    return(ses);
  }
  
  myvars = &ses->variables;
  mycolors = &ses->colors;

  arg = get_arg_stop_spaces(arg, left);
  arg = get_arg_with_spaces(arg, right);

  if(!*left) {
    ses->print((const uchar *)"#THESE COLOR VARIABLES HAS BEEN DEFINED:\n\r");
    show_list(*mycolors, ses);
 } else if(!*right) {
    for(found = FALSE,ln = searchnode_list2(*mycolors, left); ln;
	ln = searchnode_list2(ln->get_next(), left)) {
      found = TRUE;
      ln->show(ses);
    }
    if(!found)
      ses->print((const uchar *)"#THAT COLOR VARIABLE ISN'T DEFINED.\n\r");
  } else {
    text_to_color(right, buf, ses);

    if(*buf) {
      if((ln = searchnode_list(*myvars, left)) != NULL)
        delete ln;
      
      if((ln = searchnode_list(*mycolors, left)) != NULL)
        delete ln;
      ln = new listnode(mycolors, left, buf, TYPE_COLOR, 5);
      if(IS_SET(ses->message_flags, MESSAGE_VAR))
	ses->print((const uchar *)"#OK. COLOR VARIABLE DEFINED.\n\r");
    }
  }
  return(ses);
}

void color_to_text(uchar *color, uchar *text, unsigned short colmode)
{
  unsigned char buf[10], *tmp, buf2[10], buf3[BUFFER_SIZE];
  unsigned short tmp_colormode;
  int pos, spos, colornum;
  
  pos = 0;
  
  while(*color && (*color != ESC || *(color + 1) != '['))
    color++;
  
  while(*color) {
    assert(*color == ESC && *(color + 1) == '[');
    color += 2;
    
    spos = pos;
    
    while(isdigit(*color) || *color == 'm') {

      tmp = buf;
      if(*color == 'm') {
	*tmp++ = '0';
      } else {
	while(isdigit(*color))
	  *tmp++ = *color++;
      }
      *tmp = '\0';
      
      if(*color == ';' || *color == 'm') {
	colornum = atoi((const char *)buf);
	if(pos > 0)
	  pos = str_cat(text, pos, BUFFER_SIZE, (const uchar *)" ");
	
	if(colmode && colornum < 70) {
	  sprintf((char *)buf2, "\033[%dm", colornum);
	  pos = str_cat(text, pos, BUFFER_SIZE, buf2);
	  tmp_colormode = colmode;
	  color_to_bits(buf2, tmp_colormode, 0);
	}
	pos = str_cat(text, pos, BUFFER_SIZE, colors_str[colornum]);
	if(colmode && colornum < 70) {
	  if(bits_to_color(tmp_colormode, colmode, buf3) > 0)
	    pos = str_cat(text, pos, BUFFER_SIZE, buf3);
	}
      }
      if(*color == 'm') {
	spos = pos;
	color++;
      } else if(*color == ';')
	color++;
      else
	pos = spos;
    }
    while(*color && (*color != ESC || *(color + 1) != '['))
      color++;
  }
}

void text_to_color(const unsigned char *text, unsigned char *color,
		   struct session *ses)
{
  unsigned char buf[BUFFER_SIZE], *tmp, buf2[BUFFER_SIZE];
  int pos, multi, i;

  pos = str_cat(color, 0, BUFFER_SIZE, (const uchar *)"\033[");
  
  multi = FALSE;
  text = space_out(text);
  while(*text) {
    tmp = buf;
    while(*text && *text != ' ')
      *tmp++ = *text++;
    *tmp = '\0';
    
    for(i = 0; i < COLOR_NUM; i++)
      if(strcmp((const char *)colors_str[i], (const char *)buf) == 0)
        break;
    
    if(i == COLOR_NUM) {
      sprintf((char *)buf2, "#NO SUCH COLOR (%s).\n\r", buf);
      ses->print(buf2);
    } else {
      if(multi)
        pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)";");
      else
        multi = TRUE;
      
      sprintf((char *)buf2, "%d", i);
      pos = str_cat(color, pos, BUFFER_SIZE, buf2);
    }
    text = space_out(text);
  }
  
  pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)"m");
  if(!multi)
    *color = '\0';
}

void color_to_bits(const unsigned char *color, unsigned short &bits,
		   window *win)
{
  unsigned char buf[10], *tmp;
  int colornum;


  while(*color && (*color != ESC || *(color + 1) != '['))
    color++;
  
  while(*color) {
    assert(*color == ESC && *(color + 1) == '[');
    color += 2;  /* remove esc and [ */
  
    while(isdigit(*color) || *color == 'm') {
      
      tmp = buf;
      if(*color == 'm') {
	*tmp++ = '0';
      } else {
	while(isdigit(*color))
	  *tmp++ = *color++;
      }
      *tmp = '\0';
      
      if(*color == ';' || *color == 'm') {
	colornum = atoi((const char *)buf);
	
	if(colornum == 0) {
	  bits = default_color;
	} else {
	  bits &= ~COLBIT_RESET;
	  switch(colornum) {
	  case 1:
	    bits |= COLBIT_BOLD;
	    bits &= ~COLBIT_FAINT;
	    break;
	  case 2:
	    bits |= COLBIT_FAINT;
	    bits &= ~COLBIT_BOLD;
	    break;
	  case 3:
	    bits |= COLBIT_ITALICED;
	    break;
	  case 4:
	    bits |= COLBIT_UNDERLINED;
	    break;
	  case 5:
	    bits |= COLBIT_BLINK;
	    break;
	  case 7:
	    bits |= COLBIT_NEGATIVE;
	    break;
	  case 22:
	    bits &= ~COLBIT_BOLD;
	    break;
	  case 23:
	    bits &= ~COLBIT_ITALICED;
	    break;
	  case 24:
	    bits &= ~COLBIT_UNDERLINED;
	    break;
	  case 25:
	    bits &= ~COLBIT_BLINK;
	    break;
	  case 27:
	    bits &= ~COLBIT_NEGATIVE;
	    break;
	  case 30:
	  case 31:
	  case 32:
	  case 33:
	  case 34:
	  case 35:
	  case 36:
	  case 37:
	    bits = (bits & 0xf0ff) | ((colornum - 30) << 8);
	    break;
	  case 40:
	  case 41:
	  case 42:
	  case 43:
	  case 44:
	  case 45:
	  case 46:
	  case 47:
	    bits = (bits & 0x0fff) | ((colornum - 40) << 12);
	    break;
	  case 50: /* xwindow special text colors */
	  case 51:
	  case 52:
	  case 53:
	  case 54:
	  case 55:
	    bits = (bits & 0xf0ff) | ((colornum - 50 + XCOL1) << 8);
	    break;
	  case 60: /* xwindow special background colors */
	  case 61:
	  case 62:
	  case 63:
	  case 64:
	  case 65:
	    bits = (bits & 0x0fff) | ((colornum - 60 + XCOL1) << 12);
	    break;
	  case 70:
	    assert(win);
	    win->set_savecolor(bits);
	    break;
	  case 71:
	    assert(win);
	    bits = win->get_savecolor();
	    break;
	  default:
	    break;
	  }
	}
      }
      if(*color == 'm')
	break;

      if(*color == ';')
	color++;
    }
    while(*color && (*color != ESC || *(color + 1) != '['))
      color++;
  }
}

int bits_to_color(const unsigned short obits, const unsigned short nbits,
		  unsigned char *color)
{
  unsigned char tmp[20];
  unsigned short bits;
  int pos = 0;
  
  *color = '\0';
  if(obits == nbits)
    return(0);
  
  pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)"\033[");
  if(nbits & COLBIT_RESET) {
    pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)"0m");
  } else {
    if((master_table.is_set_flag(MASTER_FLAG_NO_NEG_OFF) &&
	!(nbits & COLBIT_NEGATIVE) && (obits & COLBIT_NEGATIVE)) ||
       (!(nbits & COLBIT_FAINT) && (obits & COLBIT_FAINT))) {
      pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)"0;");
      bits = 0;
    } else {
      bits = obits;
    }
    
    if((bits & 0x00ff) != (nbits & 0x00ff)) {
      // bold
      if((nbits&COLBIT_BOLD) && !(bits&COLBIT_BOLD))
	pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)"1;");
      if(!(nbits&COLBIT_BOLD) && (bits&COLBIT_BOLD))
	pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)"22;");
      
      // faint
      if((nbits&COLBIT_FAINT) && !(bits&COLBIT_FAINT))
	pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)"2;");
      
      // italiced
      if((nbits&COLBIT_ITALICED) && !(bits&COLBIT_ITALICED))
	pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)"3;");
      if(!(nbits&COLBIT_ITALICED) && (bits&COLBIT_ITALICED))
	pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)"23;");
      
      // underlined
      if((nbits&COLBIT_UNDERLINED) && !(bits&COLBIT_UNDERLINED))
	pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)"4;");
      if(!(nbits&COLBIT_UNDERLINED) && (bits&COLBIT_UNDERLINED))
	pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)"24;");
      
      // blink
      if((nbits&COLBIT_BLINK) && !(bits&COLBIT_BLINK))
	pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)"5;");
      if(!(nbits&COLBIT_BLINK) && (bits&COLBIT_BLINK))
	pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)"25;");
      
      // negative
      if((nbits&COLBIT_NEGATIVE) && !(bits&COLBIT_NEGATIVE))
	pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)"7;");
      if(!(nbits&COLBIT_NEGATIVE) && (bits&COLBIT_NEGATIVE))
	pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)"27;");
    }
    // text color
    if((bits & 0x0f00) != (nbits & 0x0f00)) {
      sprintf((char *)tmp, "%d", color_nums[(int)MODE_TO_T(nbits)]);
      pos = str_cat(color, pos, BUFFER_SIZE, tmp);
      pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)";");
    }
    // background color
    if((bits & 0xf000) != (nbits & 0xf000)) {
      sprintf((char *)tmp, "%d", color_nums[(int)MODE_TO_B(nbits)] + 10);
      pos = str_cat(color, pos, BUFFER_SIZE, tmp);
      pos = str_cat(color, pos, BUFFER_SIZE, (const uchar *)"m");
    } else {
      pos = str_cat(color, pos - 1, BUFFER_SIZE, (const uchar *)"m");
    }
  }
  return(pos);
}
