/*********************************************************************/
/* file: session.c.c - funtions related to sessions                  */
/*                             HUT TINTIN                            */
/*                coded by Ossi Tulonen from TinTin III              */
/*                TinTin III coded by peter unold 1992               */
/*********************************************************************/
/*
 * TinTin III code Copyright (c) 1992 Peter Unold
 * All changed code from TinTin III Copyright (c) 1992 1993 1994 Ossi tulonen
 * All rights reserved.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */
 
#include "structs.h"

#include <ctype.h>
#include <string.h>
#include <time.h>
#include <sys/errno.h>
#include <assert.h>

#include "comms_help.h"
#include "files.h"
#include "llist.h"
#include "main.h"
#include "net.h"
#include "parse.h"
#include "screen.h"
#include "session.h"
#ifdef TERM
#  include "term.h"
#endif
#include "utility.h"
#include "utils.h"

#ifndef X_DISPLAY_MISSING
extern XtAppContext MainContext;

#define NoID (XtInputId)-1

#  include "xscreen.h"
#endif

#ifdef TERM
#  define CLOSE(sock)                    term_close(sock)
#else
#  define CLOSE(sock)                    close(sock)
#endif

session *session::sessions = 0;

extern int quietmode;
extern struct master_table_data master_table;

extern const uchar *CEL;

extern int sys_nerr, errno;
#ifndef HAVE_SYS_ERRLIST
extern char *sys_errlist[];
#endif

// ***********************
// class session functions
// ***********************

session::session()
{
  int i;
  
  name    = strdup((const uchar *)"master");
  host    = strdup((const uchar *)"not connected");
  ip_host = 0;
  port    = 0;
  
  word_list_pos  = 0;
  word_list_size = 0;
  word_list      = 0;

  tick_size    = 75;
  sec_to_tick  = 75;
  start_time   = time(0);
  speedlimit   = -1;
  echo_flags   = DEFAULT_ECHO;
  message_flags= DEFAULT_MESS;
  ignore_flags = 0;
  logfile      = 0;
  aliases      = 0;
  actions      = 0;
  colors       = 0;
  variables    = 0;
  path_mark    = 0;
  path_length  = 0;
  path_lastnum = 0;
  path_lastdir = (uchar)' ';
  for(i = 0; i < NTELOPTS; i++) {
    will_status[i] = 0;
    do_status[i] = 0;
  }
  session_hist_size = DEFAULT_SES_HIST_SIZE;
  if(!(history = new uchar *[session_hist_size])) {
    fprintf(stderr, "Out of memory.\n");
    exit(1);
  }
  for(i = 0; i < session_hist_size; i++)
    history[i] = 0;
  
  session_hist_end = 0;
  session_hist_num = 0;
  session_hist_mark_pos = 0;
  session_hist_mark_col = 1;
  current_line[0] = '\0';
  current_pos     = 0;
  
  set_prompt((const uchar *)"> ");
  
  waitcomms    = 0;
  net          = -1;
#ifndef X_DISPLAY_MISSING
  NetId        = NoID;
  ConId        = NoID;
  output_head  = 0;
  output_tail  = 0;
#endif
  input        = window::get_active_win();
  output       = window::get_active_win();
  input->set_active_ses(this);

  ses_flags    = DEFAULT_SES_FLAGS;

  next         = 0;
}

#ifndef X_DISPLAY_MISSING
inline void XReadNet(XtPointer point, int *, XtInputId *)
{
  ReadNet((session *)point);
  DrawFunc();
}
#endif

session::session(const unsigned char *nam)
{
  int i;
  
  name    = strdup(nam);
  host    = strdup((const uchar *)"not connected");
  ip_host = 0;
  port    = 0;
  
  word_list_pos  = 0;
  word_list_size = 0;
  word_list      = 0;

  if(master_table.main_session->word_list) {
    word_list_size = master_table.main_session->word_list_size;
    word_list_pos  = master_table.main_session->word_list_pos;
    word_list      = new uchar *[word_list_size];
    
    for(i = 0; i < word_list_pos; i++) {
      word_list[i] = strdup(master_table.main_session->word_list[i]);
    }
  }
  
  tick_size    = 75;
  sec_to_tick  = 75;
  start_time   = time(0);
  speedlimit   = -1;
  echo_flags   = master_table.main_session->echo_flags;
  message_flags= master_table.main_session->message_flags;
  ignore_flags = master_table.main_session->ignore_flags;
  logfile      = 0;
  copy_list(master_table.main_session->aliases,   &aliases);
  copy_list(master_table.main_session->actions,   &actions);
  copy_list(master_table.main_session->colors,    &colors);
  copy_list(master_table.main_session->variables, &variables);
  path_mark    = 0;
  path_length  = 0;
  path_lastnum = 0;
  path_lastdir = (uchar)' ';
  for(i = 0; i < NTELOPTS; i++) {
    will_status[i] = 0;
    do_status[i] = 0;
  }

  session_hist_size = DEFAULT_SES_HIST_SIZE;
  if(!(history = new uchar *[session_hist_size])) {
    fprintf(stderr, "Out of memory.\n");
    exit(1);
  }
  for(i = 0; i < session_hist_size; i++)
    history[i] = 0;
  
  session_hist_end = 0;
  session_hist_num = 0;
  session_hist_mark_pos = 0;
  session_hist_mark_col = 1;
  current_line[0] = '\0';
  current_pos     = 0;
  
  set_prompt((const uchar *)"");
  
  waitcomms    = 0;
  net          = -1;
#ifndef X_DISPLAY_MISSING
  NetId        = NoID;
  ConId        = NoID;
  output_head  = 0;
  output_tail  = 0;
#endif
  input        = window::get_active_win();
  output       = window::get_active_win();
  
  ses_flags    = DEFAULT_SES_FLAGS;
  
  next              = session::sessions;
  session::sessions = this;
  input->set_active_ses(this);
  master_table.main_session->next = this;
}

session::~session()
{
  unsigned char buf[BUFFER_SIZE];
  int i;
  session *sesptr;
  struct waitcommdata *tmp, *tmpnext;
#ifndef X_DISPLAY_MISSING
  struct output_stack_elem *in_st, *next_in_st;
#endif
  session *newactiv;
  window *win;
  basewindow *base;
  stack *st;
  
  for(st = stack::get_all(); st; st = st->get_next())
    if(this == (const session *)st->data)
      st->data = next;
  
  sprintf((char *)buf, "#SESSION '%s' DIED.\n\r", name);
  print(buf);
  
  if(this == session::sessions)
    session::sessions = next;
  else {
    if((sesptr = session::sessions)) {
      while(sesptr->next && sesptr->next != this)
	sesptr = sesptr->next;
      
      if(sesptr->next)
	sesptr->next = next;
    }
  }
  
  for(base = basewindow::get_bases(); base; base = base->get_next()) {
    for(win = base->get_windows(); win; win = win->get_next()) {
      if(win->get_hist_on() == this) {
	win->exit_win_history();
      }
    }
  }

  if(net >= 0) {
#ifndef X_DISPLAY_MISSING
    if(NetId != NoID)
      XtRemoveInput(NetId);
    if(ConId != NoID)
      XtRemoveInput(ConId);
    
    for(in_st = output_head; in_st; in_st = next_in_st) {
      next_in_st = in_st->next;
      
      if(in_st->comm_line)
	free(in_st->comm_line);
      free(in_st);
    }
    output_head = 0;
    output_tail = 0;
#endif
    if(CLOSE(net) < 0)
      perror("#close in cleanup_session");
  }
  
  for(i = 0; i < word_list_pos; i++)
    free(word_list[i]);
  delete[] word_list;
  word_list = 0;
  
  for(i = 0; i < session_hist_size; i++)
    if(history[i])
      free(history[i]);
  delete[] history;
  history = 0;
  
  kill_list(&aliases);
  kill_list(&actions);
  kill_list(&colors);
  kill_list(&variables);
  
  for(tmp = waitcomms; tmp; tmp = tmpnext) {
    tmpnext = tmp->next;
    free(tmp->comm);
    free(tmp);
  }
  
  if(name)
    free(name);
  
  if(host)
    free(host);
  
  if(ip_host)
    free(ip_host);
  
  if(logfile)
    fclose(logfile);
  
  if(this == master_table.main_session) {
    master_table.main_session = 0;
    exit_tintin(0);
  } else {
    if(master_table.main_session)
      master_table.main_session->next = session::sessions;
  }
  
  if(this == input->get_active_ses()) {
    for(newactiv = session::get_sessions();
	newactiv;
	newactiv = newactiv->get_next()) {
      if(newactiv->get_input() == input) {
	break;
      }
    }
    if(!newactiv) {
      if(master_table.main_session) {
	newactiv = master_table.main_session;
	if(newactiv->get_input()->get_active_ses() == newactiv)
	  newactiv->get_input()->check_session();
	
	input->set_active_ses(newactiv);
	newactiv->set_input(input);
	if(!window::get_active_win()->get_active_ses()) {
	  window::get_active_win()->next_window_active(input);
	}
      }
    } else {
      input->set_active_ses(newactiv);
    }
    if(newactiv) {
      sprintf((char *)buf, "#SESSION '%s' ACTIVATED.\n\r", newactiv->get_name());
      newactiv->print(buf);
      newactiv->get_input()->set_ptexrem();
    }
  }
  input->set_ptexrem();
}

void session::rename(const uchar *new_name)
{
  if(name)
    free(name);
  name = strdup(new_name);
}

void session::set_hist_size(const int new_size)
{
  int i, pos, trans, old_size, old_end, old_num;
  uchar **old_hist;
  window *win;
  basewindow *base;
  
  if(new_size == session_hist_size)
    return;
  
  for(base = basewindow::get_bases(); base; base = base->get_next()) {
    for(win = base->get_windows(); win; win = win->get_next()) {
      if(win->get_hist_on() == this)
	win->exit_win_history();
    }
  }
  
  old_hist = history;
  old_size = session_hist_size;
  old_end  = session_hist_end;
  old_num  = session_hist_num;
  
  session_hist_size = new_size;
  if(!(history = new uchar *[session_hist_size])) {
    fprintf(stderr, "Out of memory.\n");
    exit(1);
  }
  for(i = 0; i < session_hist_size; i++)
    history[i] = 0;
  
  session_hist_mark_pos = 0;
  session_hist_mark_col = 1;
  
  trans = (new_size < old_num ? new_size : old_num);
  pos = old_end - trans;
  if(pos < 0)
    pos += old_size;
  
  for(i = 0; i < trans; i++, pos++) {
    if(pos == old_size)
      pos = 0;
    
    history[i] = old_hist[pos];
    old_hist[pos] = 0;
  }
  
  session_hist_num = trans;
  session_hist_end = trans;
  if(session_hist_end == session_hist_size)
    session_hist_end = 0;
  
  for(i = 0; i < old_size; i++)
    if(old_hist[i])
      free(old_hist[i]);
  delete[] old_hist;
}

void session::set_net(const int sock)
{
  net   = sock;
#ifndef X_DISPLAY_MISSING
  NetId = XtAppAddInput(MainContext, sock, (XtPointer)XtInputReadMask,
			XReadNet, (XtPointer)this);
#endif
}

void session::rem_net()
{
  int i;
  
  if(net >= 0) {
#ifndef X_DISPLAY_MISSING
    assert(NetId != NoID);
    
    XtRemoveInput(NetId);
#endif
    if(CLOSE(net) < 0)
      perror("#close in session::rem_net()");
  }
  net   = -1;
#ifndef X_DISPLAY_MISSING
  NetId = NoID;
#endif
  
  set_addr(0, 0, 0);
  
  print((const uchar *)"#Connection closed.\n\r> ");
  
  set_prompt((const uchar *)"> ");
  
  for(i = 0; i < NTELOPTS; i++) {
    will_status[i] = 0;
    do_status[i] = 0;
  }
  
  ses_flags = DEFAULT_SES_FLAGS;
}

#ifndef X_DISPLAY_MISSING
void session::set_ConNet(const int sock)
{
  net   = sock;
  ConId = XtAppAddInput(MainContext, sock, (XtPointer)XtInputReadMask,
			XConNet, (XtPointer)this);
  
  assert(NetId == NoID);
  
  /* XtInputExceptMask ? */
}

void session::rem_ConNet()
{
  assert(ConId != NoID);
  
  XtRemoveInput(ConId);
  ConId = NoID;
  net   = -1;
}

void session::add_output_stack(const uchar *com)
{
  output_stack_elem *tmp;
  
  CREATE(tmp, struct output_stack_elem, 1);
  
  tmp->comm_line = strdup(com);
  tmp->next      = 0;
  
  if(!output_head) {
    output_head = tmp;
  } else {
    assert(output_tail);
    
    output_tail->next = tmp;
  }
  output_tail = tmp;
}

void session::clear_output_stack()
{
  output_stack_elem *tmp, *next_tmp;
  
  for(tmp = output_head; tmp; tmp = next_tmp) {
    next_tmp = tmp->next;
    
    if(tmp->comm_line)
      free(tmp->comm_line);
    free(tmp);
  }
  output_head = 0;
  output_tail = 0;
}
#endif

void session::print(const unsigned char *text)
{
  unsigned char buf[BUFFER_SIZE];
  int pos;
  window *win;
  
  if(quietmode) {
    return;
  }
  
  if(*text) {
    insert_ses_scrhist(text, FALSE);
    
    if(master_table.output_window) {
      if(master_table.output_window == NULL_WIN)
	return;
      
      win = master_table.output_window;
    } else
      win = output;
    
    if(!win->get_hist_on()) {
      if((this == win->get_active_ses()) ||
	 (!win->get_active_ses() &&
	  window::get_active_win()->get_active_ses() == this)) {
	win->remove_user_text();
	win->print(text);
      } else if(!win->get_active_ses() || is_set_flag(SES_FLAG_SNOOP)) {
	win->remove_user_text();
	pos = 0;
	pos = str_cat(buf, pos, sizeof(buf), (const uchar *)"\r");
	pos = str_cat(buf, pos, sizeof(buf), CEL);
	win->print(buf);
	pos = 0;
	pos = str_cat(buf, pos, sizeof(buf), name);
	pos = str_cat(buf, pos, sizeof(buf), (const uchar *)"% ");
	win->print(text, buf);
	win->set_ptexrem();
	win->print_prompt(TRUE, FALSE);
      }
    }
  }
}

void session::print_prompt(const int all)
{
  if(this == output->get_active_ses()) {
    output->print_prompt(all);
  }
}

void session::set_addr(const uchar *new_host, const uchar *new_ip_host,
		       const int new_port)
{
  if(host)
    free(host);
  if(ip_host)
    free(ip_host);
  
  if(new_host)
    host = strdup(new_host);
  else
    host = strdup((const uchar *)"not connected");
  
  if(new_ip_host)
    ip_host = strdup(new_ip_host);
  else
    ip_host = 0;
  
  port = new_port;
}

const uchar *session::get_prompt() const
{
  static uchar prompt_store[MAX_PROMPT_SIZE + 2];
  int pos;
  
  if(master_table.inputbuf2_pos) {
    pos = str_cat(prompt_store, 0, sizeof(prompt_store), promptbuf);
    pos = str_cat(prompt_store, pos, sizeof(prompt_store), (const uchar *)"+ ");
    return(prompt_store);
  }
  return(promptbuf);
}

const int session::get_prompt_len() const
{
  if(master_table.inputbuf2_pos) {
    return(prompt_pos + 2);
  }
  return prompt_pos;
}

void session::set_prompt(const unsigned char *text)
{
  str_cat(promptbuf, 0, MAX_PROMPT_SIZE, text);
  prompt_pos = str_len(promptbuf);
}

void session::add_prompt(const unsigned char *text)
{
  str_cat(promptbuf, strlen((const char *)promptbuf), MAX_PROMPT_SIZE, text);
  prompt_pos = str_len(promptbuf);
}

const unsigned char *session::search_word(const unsigned char *word) const
{
  int pos;
  
  for(pos = 0; pos < word_list_pos; pos++) {
    if(strcmp((const char *)word, (const char *)word_list[pos]) == 0)
      return(word_list[pos]);
  }
  
  return(0);
}

/************************/
/* the #session command */
/************************/
session *session_command(const unsigned char *arg, session *ses)
{
  unsigned char left[BUFFER_SIZE], right[BUFFER_SIZE];
  session *newses;
  const session *sesptr;
  window *win;
  static const uchar *mean_text = (const uchar *)""
"   Name   Address                      Messages           Echoed     Ignore\n\r"
"                                       A A M P S S V W W  A A C I W  A A A C S V\n\r"
"                                       C L A A E U A A O  C L M F A  C L L A U A\n\r"
"                                       T I C T T B R I R  T I P   I  T I L L B R\n\r";
  
  arg = get_arg_stop_spaces(arg, left);
  arg = get_arg_with_spaces(arg, right);
  
  if(!*left) {
    ses->print((const uchar *)"#THESE SESSIONS HAS BEEN DEFINED:\n\r");
    ses->print(mean_text);
    for(sesptr = master_table.main_session;
	sesptr;
	sesptr = sesptr->get_next())
      show_session(sesptr, ses);

  } else if(*left && !*right) {
    for(sesptr = master_table.main_session;
	sesptr;
	sesptr = sesptr->get_next())
      if(mystrcmp(sesptr->get_name(), left))
	break;
    
    if(sesptr) {
      ses->print(mean_text);
      show_session(sesptr, ses);
      sesptr = sesptr->get_next();
      
      while(sesptr) {
	if(mystrcmp(sesptr->get_name(), left))
	  show_session(sesptr, ses);
	sesptr = sesptr->get_next();
      }
    } else {
      
      if((win = find_window(left))) {
	ses->print((const uchar *)"#THERE'S A WINDOW WITH THAT NAME ALREADY.\n\r");
	return(ses);
      }

      ses->print((const uchar *)"#OK. NEW SESSION OPENED.\n\r");
      ses = new session(left);
      ses->set_prompt((const uchar *)"> ");
    }
    
  } else {
    win = find_window(left);
    if(win) {
      ses->print((const uchar *)"#THERE'S A WINDOW WITH THAT NAME ALREADY.\n\r");
      return 0;
    }
    
    for(newses = master_table.main_session;
	newses;
	newses = newses->get_next()) {
      if(strcmp((const char *)newses->get_name(), (const char *)left) == 0) {
	if(newses->get_socket() >= 0) {
	  ses->print((const uchar *)"#THERE'S A SESSION WITH THAT NAME WHICH IS CONNECTED ALREADY.\n\r");
	  return 0;
	}
	break;
      }
    }
    
    if(!newses)
      newses = new session(left);
    
    ses = comm_connect(right, newses);
    
    if(ses->get_socket() < 0) {
      delete ses;
      return(0);
    }
  }
  
  return ses;
}

/******************/
/* show a session */
/******************/
void show_session(const session *show, session *ses) 
{
  unsigned char buf[BUFFER_SIZE], tmp[100];
  unsigned char len;

  sprintf((char *)tmp, "%c%c %-6.6s ",
	  show == window::get_active_win()->get_active_ses() ? '*' : ' ',
	  (show->is_set_flag(SES_FLAG_SNOOP) ? 'S' : ' '),
	  show->get_name());
  len = str_cat(buf, 0, sizeof(buf), tmp);
  sprintf((char *)tmp, "%-22.22s %-5d ", show->get_host(), show->get_port());
  len = str_cat(buf, len, sizeof(buf), tmp);
  
  sprintf((char *)tmp, "%c %c %c %c %c %c %c %c %c  ",
	  (IS_SET(show->message_flags, MESSAGE_ACTION) ? '*' : ' '),
	  (IS_SET(show->message_flags, MESSAGE_ALIAS) ?  '*' : ' '),
	  (IS_SET(show->message_flags, MESSAGE_MACRO) ?  '*' : ' '),
	  (IS_SET(show->message_flags, MESSAGE_PATH) ?   '*' : ' '),
	  (IS_SET(show->message_flags, MESSAGE_SET) ?    '*' : ' '),
	  (IS_SET(show->message_flags, MESSAGE_SUBS) ?   '*' : ' '),
	  (IS_SET(show->message_flags, MESSAGE_VAR) ?    '*' : ' '),
	  (IS_SET(show->message_flags, MESSAGE_WAIT) ?   '*' : ' '),
	  (IS_SET(show->message_flags, MESSAGE_WORD) ?   '*' : ' '));
  len = str_cat(buf, len, sizeof(buf), tmp);
  
  sprintf((char *)tmp, "%c %c %c %c %c  ",
	  (IS_SET(show->echo_flags, ECHO_ACTION) ? '*' : ' '),
	  (IS_SET(show->echo_flags, ECHO_ALIAS) ?  '*' : ' '),
	  (IS_SET(show->echo_flags, ECHO_CMP) ?    '*' : ' '),
	  (IS_SET(show->echo_flags, ECHO_IF) ?     '*' : ' '),
	  (IS_SET(show->echo_flags, ECHO_WAIT) ?   '*' : ' '));
  len = str_cat(buf, len, sizeof(buf), tmp);
  
  sprintf((char *)tmp, "%c %c %c %c %c %c\n\r",
	  (IS_SET(show->ignore_flags, IGNORE_ACTION) ? '*' : ' '),
	  (IS_SET(show->ignore_flags, IGNORE_ALIAS) ?  '*' : ' '),
	  (IS_SET(show->ignore_flags, IGNORE_ALL) ?    '*' : ' '),
	  (IS_SET(show->ignore_flags, IGNORE_CALC) ?   '*' : ' '),
	  (IS_SET(show->ignore_flags, IGNORE_SUBS) ?   '*' : ' '),
	  (IS_SET(show->ignore_flags, IGNORE_VAR) ?    '*' : ' '));
  len = str_cat(buf, len, sizeof(buf), tmp);
  
  if(show->logfile || show->speedlimit > 0) {
    len = str_cat(buf, len, sizeof(buf), (const uchar *)"         ");
    
    if(show->logfile)
      len = str_cat(buf, len, sizeof(buf), (const uchar *)" (logging)");
    if(show->speedlimit > 0) {
      sprintf((char *)tmp, "%d", show->speedlimit);
      len = str_cat(buf, len, sizeof(buf), (const uchar *)" (Speedwalk limit: ");
      len = str_cat(buf, len, sizeof(buf), tmp);
      len = str_cat(buf, len, sizeof(buf), (const uchar *)")");
    }
    len = str_cat(buf, len, sizeof(buf), (const uchar *)"\n\r");
  }
  ses->print(buf);
}

// *********************
//  command rename code
// *********************

session *comm_rename(const uchar *arg, session *ses)
{
  unsigned char name[BUFFER_SIZE];
  const session *sesptr;

  arg = get_arg_with_spaces(arg, name);
  if(*name) {
    if(find_window(name)) {
      ses->print((const uchar *)"#Already that named window\n\r");
      ses->set_status(FALSE);
      return(ses);
    }
    for(sesptr = master_table.main_session;
	sesptr;
	sesptr = sesptr->get_next())
      if(strcmp((const char *)sesptr->get_name(), (const char *)name) == 0) {
	ses->print((const uchar *)"#Already that named session.\n\r");
	ses->set_status(FALSE);
	return(ses);
      }
    
    ses->rename(name);
    ses->print((const uchar *)"#Ok.\n\r");
    ses->set_status(TRUE);
  } else {
    ses->print(help_rename);
    ses->set_status(FALSE);
  }
  return(ses);
}

// *********************
//  command rename code
// *********************

session *comm_sethistsize(const uchar *arg, session *ses)
{
  uchar size_buf[BUFFER_SIZE];
  int size;
  
  arg = get_arg_with_spaces(arg, size_buf);
  if(*size_buf) {
    size = string_to_number(size_buf, ses);
    
    if(size < 10 || size > 32000) {
      ses->print((const uchar *)"#Size must be 10-3200.\n\r");
    } else {
      ses->set_hist_size(size);
      ses->print((const uchar *)"#Ok.\n\r");
    }
  } else {
    ses->print(help_sethistsize);
  }
  return(ses);
}

// *****************************
//  code to show error messages
// *****************************

void session::print_errmsg(const char *text)
{
  uchar buffer[BUFFER_SIZE];
  int pos;
  
  pos = str_cat(buffer, 0, sizeof(buffer), (const uchar *)"#ERROR: ");
  pos = str_cat(buffer, pos, sizeof(buffer), (const uchar *)text);
  pos = str_cat(buffer, pos, sizeof(buffer), (const uchar *)" ");
  if(errno > 0 && errno < sys_nerr) {
    pos = str_cat(buffer, pos, sizeof(buffer),
		  (const uchar *)sys_errlist[errno]);
  } else {
    pos = str_cat(buffer, pos, sizeof(buffer), (const uchar *)"(Unknown)");
  }
  pos = str_cat(buffer, pos, sizeof(buffer), (const uchar *)"\n\r");
  print(buffer);
}
