#ifndef __SESSION_H__
#define __SESSION_H__ 1

#define SES_FLAG_EMAILCHECK   (1L<<0)        // Email check on/off
#define SES_FLAG_KEYBELL      (1L<<1)        // key bell on/off
#define SES_FLAG_LINEMODE     (1L<<2)        // linemode
#define SES_FLAG_SNOOP        (1L<<3)        // session snooped
#define SES_FLAG_TERMECHO     (1L<<4)        // terminal echoing
#define SES_FLAG_TICKER       (1L<<5)        // ticker on/off
#define SES_FLAG_WFESPACE     (1L<<6)        // wordfill extra space
#define SES_FLAG_VERBATIMMODE (1L<<7)        // verbatimmode on/off
#define SES_FLAG_GA_PROMPT    (1L<<8)        // prompt's must end with IAC GA
#define SES_FLAG_ASK_CONFIRM  (1L<<9)        // asking confirmation with quit
                                             // funktion.
#define SES_FLAG_SPEEDWALK    (1L<<10)       // speedwalking on/off

session *session_command(const unsigned char *arg, session *ses);
void show_session(const session *, session *);
session *comm_rename(const uchar *, session *);
session *comm_sethistsize(const uchar *, session *);

void set_mail_var(const uchar *);

// from net.cc file
void ReadNet(session *ses);
#ifndef X_DISPLAY_MISSING
struct output_stack_elem {
  uchar *comm_line;
  struct output_stack_elem *next;
};
#endif

class session {
public:
  session();
  session(const unsigned char *nam);
  ~session();

  void rename(const uchar *);
  void set_hist_size(const int);

  void set_net(const int);
  void rem_net();
#ifndef X_DISPLAY_MISSING
  void set_ConNet(const int sock);
  void rem_ConNet();
  void add_output_stack(const uchar *);
  void clear_output_stack();
  struct output_stack_elem *get_output_stack() const { return output_head; }
#endif
  
  void print(const unsigned char *text);
  void print_prompt(const int all);
  
  const unsigned char *search_word(const unsigned char *) const;
  
  // from net.cc file
friend void ReadNet(session *ses);
    
  session *write_net(const unsigned char *line);
  session *send_naws();
  session *do_telnet_protecol(int telcom, int telopt);

  
  // from history.cc file
  void session::insert_ses_scrhist(const unsigned char *text, const int flu);
  const unsigned char *search_hist_word(const unsigned char *, int &, const int ignorecase = FALSE) const;
  

  // from variable.cc file
  void set_status(const int);
  
  const int get_hist_size() const { return session_hist_size; }
  const int get_hist_num()  const { return session_hist_num; }
  const int get_hist_end()  const { return session_hist_end; }
  const unsigned char *get_scrhistory(const int pos) const {
    return history[pos];
  }
  void set_hist_mark(const int pos, const int col) {
    session_hist_mark_pos = pos;
    session_hist_mark_col = col;
  }
  const int get_hist_markp() const { return session_hist_mark_pos; }
  const int get_hist_markc() const { return session_hist_mark_col; }

  const unsigned char *get_name() const { return name; }
  const uchar *get_host() const    { return host; }
  const uchar *get_ip_host() const { return ip_host; }
  const int get_port() const       { return port; }
  void set_addr(const uchar *, const uchar *, const int);
  
  window *get_input() const             { return input; }
  void set_input(window *win)           { input = win; }
  window *get_output() const            { return output; }
  void set_output(window *win)          { output = win; }
  const time_t get_time() const         { return start_time; }

  const uchar *get_prompt() const;
  const int get_prompt_len() const;
  void set_prompt(const unsigned char *text);
  void add_prompt(const unsigned char *);

  void print_errmsg(const char *);
  
  // in telnet_opt.cc file
  int get_will_wont(int) const;
  int get_do_dont(int) const;

  int send_wont(int);

  const int get_socket() const          { return net; }
#ifndef X_DISPLAY_MISSING
  const is_trycon() const               { return(ConId != (XtInputId)-1); }
#endif
  
  // from path.cc
  void insert_path(uchar);
  const uchar remove_path();

  const int path_to_string(uchar *, const int);

  void reset_path() {
    path_mark   = 0;
    path_length = 0;
    path_lastnum = 0;
    path_lastdir = (uchar )' ';
  }
  
  void put_nextstack(stack *st) const { st->data = next; }
  session *get_next() const           { return next; }
  static session *get_sessions()      { return sessions; }

  ulong echo_flags, ignore_flags, message_flags;
  listnode *aliases, *actions, *colors, *variables;
  waitcommdata *waitcomms;
  int tick_start_time, tick_size, sec_to_tick, speedlimit;
  int word_list_pos, word_list_size;
  uchar **word_list;
  FILE *logfile;
  
  txt_q netinput;
  
  int is_set_flag(const ulong bit) const { return(ses_flags & bit); }
  void set_flag(const ulong bit)         { ses_flags |= bit; }
  void rem_flag(const ulong bit)         { ses_flags &= ~bit; }
  
private:
  unsigned char *name;
  unsigned char *host;
  unsigned char *ip_host;
  int port;
  unsigned char current_line[MAX_LINE_LENGTH], **history;
  int session_hist_size, session_hist_end, session_hist_num;
  int session_hist_mark_pos, session_hist_mark_col;
  int current_pos;
  
  unsigned char will_status[NTELOPTS];
  unsigned char do_status[NTELOPTS];
  unsigned char promptbuf[MAX_PROMPT_SIZE];
  int prompt_pos;
  int net;
#ifndef X_DISPLAY_MISSING
  XtInputId NetId;
  XtInputId ConId;
  
  struct output_stack_elem *output_head, *output_tail;
#endif
  
  time_t start_time;
  window *input, *output;
  
  unsigned char path[MAX_PATH_LENGTH], path_lastdir;
  int path_mark, path_length, path_lastnum;
  
  ulong ses_flags;
  
  static session *sessions;
  session *next;
};

#endif
