/*********************************************************************/
/* File: ticks.c - functions for the ticker stuff                    */
/*                             HUT TINTIN                            */
/*                coded by Ossi Tulonen from TinTin III              */
/*                TinTin III coded by peter unold 1992               */
/*********************************************************************/
/*
 * TinTin III code Copyright (c) 1992 Peter Unold
 * All changed code from TinTin III Copyright (c) 1992 1993 1994 Ossi tulonen
 * All rights reserved.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */
 
#include "structs.h"

#include <ctype.h>
#include <string.h>
#include <time.h>
#include <assert.h>

#include "comms_help.h"
#include "llist.h"
#include "parse.h"
#include "screen.h"
#include "session.h"
#include "ticks.h"
#include "utility.h"
#include "utils.h"

/* externs */
extern struct master_table_data master_table;
extern int default_chars_per_line;
extern int default_lines_in_scr;
extern int idletime;

/* local globals */
int window_interrupted;

/*********************/
/* the #tick command */
/*********************/
session *tick_command(const unsigned char *, session *ses)
{
  if(ses->is_set_flag(SES_FLAG_TICKER)) {
    unsigned char buf[100];
    sprintf((char *)buf, "THERE'S NOW %d SECONDS TO NEXT TICK.\n\r",
	    ses->sec_to_tick);
    ses->print(buf);
  } else
    ses->print((const uchar *)"MY TICKER IS OFF! DUNNO SECONDS TO TICK\n\r");

  return(ses);
}

/************************/
/* the #tickset command */
/************************/
session *tickset_command(const unsigned char *, session *ses)
{
  ses->tick_start_time = time(NULL);
  /* we don't prompt! too many ticksets... */
  return(ses);
}

/*************************/
/* the #ticksize command */
/*************************/
session *ticksize_command(const unsigned char *arg, session *ses)
{
  int newsize;

  if(!*arg) {
    ses->print((const uchar *)"#SET THE TICK-SIZE TO WHAT?\n\r");
    return(ses);
  }

  if(!isdigit(*arg)) {
    ses->print((const uchar *)"#SPECIFY A NUMBER!!!!TRYING TO CRASH ME EH?\n\r");
    return(ses);
  }

  newsize = atoi((const char *)arg);
  
  if(newsize != ses->tick_size) {
    ses->tick_size       = newsize;
    ses->tick_start_time = time(NULL);
    ses->print((const uchar *)"#OK NEW TICKSIZE SET\n\r");
    
  } else {
    ses->print((const uchar *)"#TICKSIZE WAS ALREADY THAT.\n\r");
  }
  
  return(ses);
}

void TickFunc()
{
  unsigned char buf[BUFFER_SIZE], *timestr, idlebuf[30];
  int oldtime;
  session *sesresult, *ses;
  struct waitcommdata *tmp, *tmpnext, *last;
  window *win;
  basewindow *base;
  listnode *ln;
  stack *st;
  
  sprintf((char *)idlebuf, "%d", idletime++);
  
  oldtime = time(0);
  
  timestr = own_asctime();
  
  st = new stack;
  
  for(ses = master_table.main_session; ses; ses = (session *)st->data) {
    ses->put_nextstack(st);
    sesresult = ses;
    
    if(ses->is_set_flag(SES_FLAG_TICKER)) {
      ses->sec_to_tick = (ses->tick_size -
			  ((oldtime - ses->tick_start_time) % 
			   ses->tick_size));
      if(ses->sec_to_tick == ses->tick_size) {
	ses->get_output()->set_ptexrem();
	ses->print((const uchar *)"#TICK!!!\n\r");
	ses->get_output()->print_prompt(TRUE);
      } else if(ses->sec_to_tick == 10) {
	ses->get_output()->set_ptexrem();
	ses->print((const uchar *)"#10 SECONDS TO TICK!!!\n\r");
	ses->get_output()->print_prompt(TRUE);
      }
    }
    if((ln = searchnode_list(ses->colors, (unsigned char *)"TIME")))
      delete ln;
    
    if((ln = searchnode_list(ses->variables, (unsigned char *)"TIME"))) {
      if(ln->right)
	free(ln->right);
      ln->right = strdup(timestr);
    } else
      ln = new listnode(&ses->variables, (unsigned char *)"TIME", timestr,
			TYPE_VAR, 5);
    
    if((ln = searchnode_list(ses->colors, (unsigned char *)"IDLETIME")))
      delete ln;
    
    if((ln = searchnode_list(ses->variables, (unsigned char *)"IDLETIME"))) {
      if(ln->right)
	free(ln->right);
      ln->right = strdup(idlebuf);
    } else
      ln = new listnode(&ses->variables, (unsigned char *)"IDLETIME", idlebuf,
			TYPE_VAR, 5);
    
    for(tmp = ses->waitcomms; tmp; tmp = tmpnext) {
      tmpnext = tmp->next;
	  tmp->time--;
      if(!tmp->time) {
	if(IS_SET(ses->echo_flags, ECHO_WAIT)) {
	  sprintf((char *)buf, "[WAIT: %s]\n\r", tmp->comm);
	  ses->print(buf);
	}
	strcpy((char *)buf, (const char *)tmp->comm);
	free(tmp->comm);
	if(tmp == ses->waitcomms) {
	  ses->waitcomms = tmp->next;
	  free(tmp);
	} else {
	  for(last = ses->waitcomms;
	      last && last->next != tmp;
	      last = last->next)
	    ;
	  if(last)
	    last->next = tmp->next;
	  else
	    assert(0);
	  free(tmp);
	}
	
	sesresult = parse_input(buf, ses);
	prompt();
	if(!sesresult)
	  break;
      }
    }
  }

  delete st;

  for(base = basewindow::get_bases(); base; base = base->get_next()) {
    for(win = base->get_windows(); win; win = win->get_next()) {
      if(win->get_delay() > 0  &&
	 win->lastupdate + win->get_delay() <= oldtime) {
	win->print_time(FALSE);
	win->lastupdate = oldtime;
      }
    }
  }
  win = window::get_active_win();
  win->move(win->get_cursor(), TRUE);
}

void window_func(...)
{
#ifndef X_DISPLAY_MISSING
  assert(!MainDisplay);
#endif

#ifndef HAVE_SIGACTION
  signal(SIGWINCH, window_func);
#endif
  
  window_interrupted = TRUE;
  get_screen_size();
}

// *************************
//  wait and unwait command
// *************************

session *comm_wait(const unsigned char *arg, session *ses)
{
  unsigned char time[BUFFER_SIZE], comm[BUFFER_SIZE];
  int num;
  struct waitcommdata *newcomm;

  arg = get_arg_stop_spaces(arg, time);
  arg = get_arg_with_spaces(arg, comm);
  if(!*time) {
    ses->print((const uchar *)"Following commands are waiting:\n\r");
    for(newcomm = ses->waitcomms; newcomm; newcomm = newcomm->next) {
      sprintf((char *)comm, "Wait time %3d: %s\n\r", newcomm->time, newcomm->comm);
      ses->print(comm);
    }
    return(ses);
  }

  num = string_to_number(time, ses);
  if(!*comm || num <= 0) {
    ses->print(help_wait);
    return(ses);
  }
  CREATE(newcomm, struct waitcommdata, 1);
  newcomm->comm = strdup(comm);
  newcomm->time = num;
  newcomm->next = ses->waitcomms;
  ses->waitcomms = newcomm;
  if(IS_SET(ses->message_flags, MESSAGE_WAIT))
    ses->print((const uchar *)"#OK. New waitcomm set.\n\r");
  return(ses);
}

session *comm_unwait(const unsigned char *arg, session *ses)
{
  unsigned char comm[BUFFER_SIZE];
  int found = 0;
  struct waitcommdata *wcom, *pre_wcom, *next_wcom;
  
  arg = get_arg_with_spaces(arg, comm);
  
  if(!*comm) {
    ses->print(help_unwait);
    return(ses);
  }
  
  while(ses->waitcomms && mystrcmp(ses->waitcomms->comm, comm)) {
    wcom = ses->waitcomms;
    ses->waitcomms = ses->waitcomms->next;
    free(wcom->comm);
    free(wcom);
    found++;
  }
  
  if(ses->waitcomms) {
    pre_wcom = ses->waitcomms;
    wcom = pre_wcom->next;
    while(wcom) {
      next_wcom = wcom->next;
      
      if(mystrcmp(wcom->comm, comm)) {
	pre_wcom->next = wcom->next;
	free(wcom->comm);
	free(wcom);
	found++;
      } else {
	pre_wcom = wcom;
      }
      wcom = next_wcom;
    }
  }
  
  if(IS_SET(ses->message_flags, MESSAGE_WAIT)) {
    if(!found)
      ses->print((const uchar *)"No such command in wait list.\n\r");
    else if(found == 1)
      ses->print((const uchar *)"#OK. Waitcomm removed.\n\r");
    else
      ses->print((const uchar *)"#OK. Waitcomms removed.\n\r");
  }
  return(ses);
}

// *****************************************
//  functions for time line string checking
// *****************************************

int check_timeline_string(const uchar *line, session *ses)
{
  uchar buf[256];
  int trans_edge = FALSE;
  
  while(*line) {
    if(*line == '%') {
      line++;
      
      if(*line == 'f') {
	if(trans_edge) {
	  ses->print((const uchar *)"# You can use only one %f code in timeline.\n\r");
	  return(FALSE);
	}
	trans_edge = TRUE;
	line++;

      } else if(*line == 'd' || // date
		*line == 't' || // time
		*line == 'b' || // bell
		*line == 'a' || // active string
		*line == 'w' || // window name
		*line == 's' || // active session name
		*line == 'o' || // sessions online time
		*line == 'T' || // ticker time to left next tick
		*line == '%') { // '%' char
	line++;

      } else {
	sprintf((char *)buf, "# Unknown %% sequence (%%%c).\n\r",
		(*line ? *line : ' '));
	ses->print(buf);
	return(FALSE);
      }
      
    } else
      line++;
  }
  
  return(TRUE);
}

/************************/
/* the #time command    */
/************************/
session *comm_time(const unsigned char *args, session *ses)
{
  uchar buf[BUFFER_SIZE];
  const uchar *newline = 0;
  int value;
  
  args = get_arg_stop_spaces(args, buf);
  
  if(*buf) {
    value = string_to_number(buf, ses);
    
    args = get_arg_stop_spaces(args, buf);
    
    if(*buf) { // New time line
      
      if(is_abrev((const uchar *)"default", buf)) {
	newline = DEF_TIME_LINE;

      } else {
	if(!check_timeline_string(buf, ses))
	  return(ses);
	
	newline = buf;
      }
    }
    
  } else
    value = -1;
  
  if(!ses->get_input()->change_time(value, newline)) {
    ses->print(help_time);
  }
  return(ses);
}
