/*********************************************************************/
/* file: variable.c - funtions related the the variable command      */
/*                            HUT TINTIN                             */
/*                     coded by Ossi Tulonen                         */
/*********************************************************************/
/*
 * Copyright (c) 1992 1993 1994 Ossi tulonen
 * All rights reserved.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */
 
#include "structs.h"

#include <ctype.h>
#include <string.h>
#include <assert.h>

#include "llist.h"
#include "parse.h"
#include "screen.h"
#include "session.h"
#include "variable.h"
#include "utility.h"
#include "utils.h"

extern struct master_table_data master_table;
extern unsigned char text_start, text_end;

/*************************/
/* the #variable command */
/*************************/
session *comm_variable(const unsigned char *arg, session *ses)
{
  unsigned char left[BUFFER_SIZE], right[BUFFER_SIZE], buf[BUFFER_SIZE];
  unsigned char word[BUFFER_SIZE], name[BUFFER_SIZE];
  const unsigned char *tmp;
  int addword = FALSE, removeword = FALSE, verbose = FALSE, first = FALSE;
  int second_var = FALSE, deletesecond = FALSE, next_is_var = FALSE;
  int pos, found;
  listnode **myvars, **mycolors, *ln;
  
  myvars = &ses->variables;
  mycolors = &ses->colors;
  
  arg = get_arg_stop_spaces(arg, left);
  while(*left == '-' && !next_is_var) {
    tmp = left + 1;
    while(isspace(*tmp))
      tmp++;
    
    while(*tmp) {
      switch(*tmp) {
      case '-':
	next_is_var = TRUE;
	break;
	
      case '2':
	second_var = TRUE;
	if(*(tmp + 1) == 'd' || *(tmp + 1) == 'D') {
	  tmp++;
	  deletesecond = TRUE;
	}
	break;
	
      case 'a':
      case 'A':
	addword = TRUE;
	break;
	
      case 'f':
      case 'F':
	first = TRUE;
	break;

      case 'r':
      case 'R':
	removeword = TRUE;
	break;

      case 'v':
      case 'V':
	verbose = TRUE;
	break;
	
      default:
	sprintf((char *)buf, "#Unknown flag '-%c' in variable command.\n\r",
		*tmp);
	ses->print(buf);
	return(ses);
      }
      tmp++;
      while(isspace(*tmp))
	tmp++;
    }
    arg = get_arg_stop_spaces(arg, left);
  }
  
  if(!second_var) {
    arg = get_arg_with_spaces(arg, word);

  } else {
    arg = get_arg_stop_spaces(arg, word);

    if(!*word) {
      ses->print((const uchar *)"#variable command with -2 flag and no second argument.\n\r");
      return(ses);
    }
    
    if(!(ln = searchnode_list(*myvars, word))) {
      ses->print((const uchar *)"#No such variable.\n\r");
      return(ses);
    }
    
    strcpy((char *)name, (const char *)word);
    strcpy((char *)word, (const char *)ln->right);
    
    if(deletesecond) {
      delete ln;
      
      if(first) { // return all but first word back to variable
	tmp = get_arg_stop_spaces(word, buf); 
	while(*tmp == ' ')
	  tmp++;
	
	ln = new listnode(myvars, name, tmp, TYPE_VAR, 5);
      }
    }
  }
  
  if(first) {
    get_arg_stop_spaces(word, right);
  } else {
    strcpy((char *)right, (const char *)word);
  }
  
  if(!*left) {
    ses->print((const uchar *)"#THESE VARIABLES HAS BEEN DEFINED:\n\r");
    show_list(*myvars, ses);
    ses->print((const uchar *)"#THESE COLOR VARIABLES HAS BEEN DEFINED:\n\r");
    show_list(*mycolors, ses);
    return(ses);
  }

  if((ln = searchnode_list(*mycolors, left)) != NULL)
    delete ln;
  
  ln = searchnode_list(*myvars, left);
  
  if(removeword || addword) {
    strcpy((char *)word, (const char *)right);
    pos = str_cat(right, 0, sizeof(right), (const uchar *)"");
    
    if(ln) {
      if(removeword || addword) {
	tmp = ln->right;
	found = FALSE;
	
	while(isspace(*tmp))
	    tmp++;
	
	while(*tmp) {
	  tmp = get_arg_stop_spaces(tmp, buf);
	  if(strcmp((const char *)word, (const char *)buf)) {
	    if(pos > 0)
	      pos = str_cat(right, pos, sizeof(right), (const uchar *)" ");
	    right[pos++] = text_start;
	    pos = str_cat(right, pos, sizeof(right) , buf);
	    right[pos++] = text_end;
	    right[pos] = '\0';
	  } else
	    found = TRUE;
	  
	  while(isspace(*tmp))
	    tmp++;
	};
	
	if(verbose && !found && !addword) {
	  ses->print((const uchar *)"#Didn't find such word in that variable.\n\r");
	}
      }
    }
    
    if(addword) {
      if(pos > 0)
	pos = str_cat(right, pos, sizeof(right), (const uchar *)" ");
      right[pos++] = text_start;
      pos = str_cat(right, pos, sizeof(right) , word);
      right[pos++] = text_end;
      right[pos] = '\0';
    }
  }
  
  if(ln)
    delete ln;
  
  ln = new listnode(myvars, left, right, TYPE_VAR, 5);
  if(IS_SET(ses->message_flags, MESSAGE_VAR) || verbose) {
    ses->print((const uchar *)"#OK. VARIABLE DEFINED.\n\r");
  }
  
  return(ses);
}

/***************************/
/* the #unvariable command */
/***************************/
session *comm_unvariable(const unsigned char *arg, session *ses)
{
  unsigned char left[BUFFER_SIZE];
  int found;
  listnode **myvars, **mycolors, *ln;
  stack *st;

  myvars = &ses->variables;
  mycolors = &ses->colors;

  arg = get_arg_stop_spaces(arg, left);

  st = new stack;

  for(found = FALSE,ln = searchnode_list2(*myvars, left); ln;
      ln = searchnode_list2((const listnode *)st->data, left)) {
    found = TRUE;
    ln->put_nextstack(st);
    delete ln;
    if(IS_SET(ses->message_flags, MESSAGE_VAR))
      ses->print((const uchar *)"#OK. VARIABLE DELETED\n\r");
  }
  for(ln = searchnode_list2(*mycolors, left); ln;
      ln = searchnode_list2((const listnode *)st->data, left)) {
    found = TRUE;
    ln->put_nextstack(st);
    delete ln;
    if(IS_SET(ses->message_flags, MESSAGE_VAR))
      ses->print((const uchar *)"#OK. COLOR VARIABLE DELETED\n\r");
  }
  if(!found)
    ses->print((const uchar *)"#THAT VARIABLE IS NOT DEFINED.\n\r");
  
  delete st;
  
  return(ses);
}

// **************************
//  status variable set code
// **************************

void session::set_status(const int value)
{
  listnode *ln;
  
  if((ln = searchnode_list(colors, (const uchar *)"STATUS")))
    delete ln;
  
  if((ln = searchnode_list(variables, (const uchar *)"STATUS"))) {
    if(ln->right)
      free(ln->right);
    if(value)
      ln->right = strdup((const uchar *)"1");
    else
      ln->right = strdup((const uchar *)"0");
  } else {
    if(value)
      ln = new listnode(&variables, (const uchar *)"STATUS",
			(const uchar *)"1", TYPE_VAR, 5);
    else
      ln = new listnode(&variables, (const uchar *)"STATUS",
			(const uchar *)"0", TYPE_VAR, 5);
  }
  if(IS_SET(message_flags, MESSAGE_VAR)) {
    print((const uchar *)"#OK. STATUS VARIABLE SET.\n\r");
  }
}

// ************************
//  mail variable set code
// ************************

void set_mail_var(const uchar *text)
{
  session *ses;
  listnode *ln;

  assert(text);
  
  for(ses = master_table.main_session; ses; ses = ses->get_next()) {
    
    if((ln = searchnode_list(ses->colors, (const uchar *)"MAIL")))
      delete ln;
    
    if((ln = searchnode_list(ses->variables, (const uchar *)"MAIL"))) {
      if(strcmp((const char *)ln->right, (const char *)text) == 0)
	continue;
      
      if(ln->right)
	free(ln->right);
      ln->right = strdup(text);
    } else {
      ln = new listnode(&ses->variables, (const uchar *)"MAIL",
			text, TYPE_VAR, 5);
    }
  }
}
