/*********************************************************************/
/* File: xscreen.cc - functions for the X-Window handling            */
/*                            HUT TINTIN                             */
/*                     coded by Ossi Tulonen                         */
/*********************************************************************/
/*
 * Copyright (c) 1995, 1996 Ossi tulonen
 * All rights reserved.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "structs.h"

#include <ctype.h>
#include <sys/time.h>
#include <assert.h>

#include <X11/IntrinsicP.h>
#include <X11/Xatom.h>

extern XtAppContext MainContext;
unsigned long colors[16];
XFontStruct *def_font = 0;

#include "main.h"
#include "screen.h"
#include "session.h"
#include "utility.h"
#include "xscreen.h"

extern struct master_table_data master_table;

extern const char *color_names[];

int button = -1, style = 0;

extern uchar *selection;
extern int selection_len;

void DrawFunc()
{
  basewindow *base;
  
  for(base = basewindow::get_bases(); base; base = base->get_next()) {
    if(base->x_wait_flush) {
      Redisplay(base->widget, 0, 0);
    }
  }
}

void Redisplay(Widget w, XEvent *, Region)
{
  int i;
  basewindow *base;

  for(base =  basewindow::get_bases(); base; base = base->get_next())
    if(base->widget == w)
      break;
  
  if(base) {
    base->x_wait_flush = FALSE;
    
    for(i = 0; i < base->rows; i++) {
      xprint_many(base, 0, i, base->cols, FALSE);
    }
    window::get_active_win()->xprint_cursor(TRUE);
    if(base->clip_color) {
      delete base->clip_color;
      base->clip_color = 0;
    }
  }
}

void Resize(Widget w)
{
  int col, row, font_width, font_height;
  basewindow *base;

  for(base = basewindow::get_bases(); base; base = base->get_next())
    if(base->widget == w)
      break;
  
  if(base) {
    font_width  = base->font->max_bounds.width;
    font_height = base->font->ascent + base->font->descent;
    col = w->core.width/font_width;
    row = w->core.height/font_height;
    
    if(row < 3)
      row = 3;
    
    if(col < 50)
      col = 50;
    
    if(base->resize(row, col) >= 0) {
      if(col * font_width != w->core.width ||
	 row * font_height != w->core.height)
	XResizeWindow(MainDisplay, XtWindow(w),
		      col * font_width, row * font_height);
    }
  }
}

// ******************
//  Visual bell code
// ******************

class OWinAndStack {
public:
  stack *st;
  const window *owin;
};

void RemXVBell(XtPointer o_tr, XtIntervalId *)
{
  int i;
  OWinAndStack *tr = (OWinAndStack *)o_tr;
  const window *win = (const window *)tr->st->data;
  
  if(tr->owin == win) { /* window is still there */
    const basewindow *base = win->get_base();
    
    for(i = win->get_scroll()->start - 1; i < win->get_scroll()->end; i++) {
      xprint_many(base, 0, i, base->cols, FALSE);
    }
    if(win == window::get_active_win())
      window::get_active_win()->xprint_cursor(TRUE);
  }
  
  delete tr->st;
  delete tr;
}

void XVBell(const window *win)
{
  int i;
  const basewindow *base = win->get_base();
  OWinAndStack *tr;
  
  tr = new OWinAndStack;
  tr->st = new stack;
  tr->st->data = (void *)win;
  tr->owin     = win;
  
  Redisplay(base->widget, 0, 0);
  
  for(i = win->get_scroll()->start - 1; i < win->get_scroll()->end; i++) {
    xprint_many(base, 0, i, base->cols, TRUE);
  }
  XtAppAddTimeOut(MainContext, 200, RemXVBell, (XtPointer)tr); // 0.2 secs
}

// **********************************
//  X-Window windows print functions
// **********************************

void xprint_one(const basewindow *base, const int col, const int row,
		const int neg)
{
  char buf[2] = { '\0', '\0'};
  int font_width, font_height, font_ascent;
  const winhist_elem *tmp;

  if(base->x_wait_flush)
    return;
  
  font_width  = base->font->max_bounds.width;
  font_height = base->font->ascent + base->font->descent;
  font_ascent = base->font->ascent;
  tmp = &base->data[row][col];

  if(neg) {
    if(tmp->colormode & COLBIT_NEGATIVE) {
      XSetBackground(MainDisplay, base->gc,
		     colors[MODE_TO_B(tmp->colormode)]);
      XSetForeground(MainDisplay, base->gc,
		     colors[MODE_TO_T(tmp->colormode)]);
    } else {
      XSetBackground(MainDisplay, base->gc,
		     colors[MODE_TO_T(tmp->colormode)]);
      XSetForeground(MainDisplay, base->gc,
		     colors[MODE_TO_B(tmp->colormode)]);
    }
  } else {
    if(tmp->colormode & COLBIT_NEGATIVE) {
      XSetBackground(MainDisplay, base->gc,
		     colors[MODE_TO_T(tmp->colormode)]);
      XSetForeground(MainDisplay, base->gc,
		     colors[MODE_TO_B(tmp->colormode)]);
    } else {
      XSetBackground(MainDisplay, base->gc,
		     colors[MODE_TO_B(tmp->colormode)]);
      XSetForeground(MainDisplay, base->gc,
		     colors[MODE_TO_T(tmp->colormode)]);
    }
  }
  buf[0] = (char)tmp->letter;
  XDrawImageString(MainDisplay, XtWindow(base->widget), base->gc,
		   col * font_width,
		   font_ascent + row * font_height,
		   (const char *)buf, 1);
  
  if(tmp->colormode & COLBIT_BOLD)
    XDrawString(MainDisplay, XtWindow(base->widget), base->gc,
		1 + col * font_width,
		font_ascent + row * font_height,
		(const char *)buf, 1);
  
  if(tmp->colormode & COLBIT_UNDERLINED) {
    XDrawLine(MainDisplay, XtWindow(base->widget), base->gc,
	      col * font_width,
	      (row + 1) * font_height - 1,
	      (col + 1) * font_width,
	      (row + 1) * font_height - 1);
  }
}

void xprint_many(const basewindow *base, const int col, const int row,
		 const int len, const int neg)
{
  uchar buffer[BUFFER_SIZE], *buf;
  int i, scol, endpos;
  int font_width, font_height, font_ascent;
  const winhist_elem *line;
  
  assert(base->cols < BUFFER_SIZE);
  assert(col + len <= base->cols);
  assert(row >= 0 && row < base->rows);
  
  if(base->x_wait_flush)
    return;
  
  endpos = col + len;
  
  font_width  = base->font->max_bounds.width;
  font_height = base->font->ascent + base->font->descent;
  font_ascent = base->font->ascent;

  line = base->data[row];
  
  i = col;
  while(i < endpos) {
    scol = i;
    buf = buffer;
    
    if(neg) {
      if(line[i].colormode & COLBIT_NEGATIVE) {
	XSetBackground(MainDisplay, base->gc,
		       colors[MODE_TO_B(line[i].colormode)]);
	XSetForeground(MainDisplay, base->gc,
		       colors[MODE_TO_T(line[i].colormode)]);
      } else {
	XSetBackground(MainDisplay, base->gc,
		       colors[MODE_TO_T(line[i].colormode)]);
	XSetForeground(MainDisplay, base->gc,
		       colors[MODE_TO_B(line[i].colormode)]);
      }
    } else {
      if(line[i].colormode & COLBIT_NEGATIVE) {
	XSetBackground(MainDisplay, base->gc,
		       colors[MODE_TO_T(line[i].colormode)]);
	XSetForeground(MainDisplay, base->gc,
		       colors[MODE_TO_B(line[i].colormode)]);
      } else {
	XSetBackground(MainDisplay, base->gc,
		       colors[MODE_TO_B(line[i].colormode)]);
	XSetForeground(MainDisplay, base->gc,
		       colors[MODE_TO_T(line[i].colormode)]);
      }
    }
    
    *buf++ = line[i++].letter;
    
    while(i < endpos && line[i].colormode == line[i - 1].colormode)
      *buf++ = line[i++].letter;
    *buf = '\0';
    
    XDrawImageString(MainDisplay, XtWindow(base->widget), base->gc,
		     scol * font_width,
		     font_ascent + row * font_height,
		     (const char *)buffer, buf - buffer);
    
    if(line[scol].colormode & COLBIT_BOLD) {
      XDrawString(MainDisplay, XtWindow(base->widget), base->gc,
		  1 + scol * font_width,
		  font_ascent + row * font_height,
		  (const char *)buffer, buf - buffer);
    }
    if(line[scol].colormode & COLBIT_UNDERLINED) {
      XDrawLine(MainDisplay, XtWindow(base->widget), base->gc,
		scol * font_width,
		font_height - 1 + row * font_height,
		(scol + buf - buffer) * font_width,
		font_height - 1 + row * font_height);
    }
  }
}

void xprint_reg(basewindow *base, const int neg)
{
  int i;
  color_clip_reg *reg;
  
  reg = base->clip_color;
  
  if(reg->s_row == reg->e_row) {
    xprint_many(base, reg->s_col, reg->s_row, reg->e_col - reg->s_col + 1,
		neg);
  } else {
    xprint_many(base, reg->s_col, reg->s_row, base->cols - reg->s_col, neg);
    
    for(i = reg->s_row + 1; i < reg->e_row; i++) {
      xprint_many(base, 0, i, base->cols, neg);
    }
    xprint_many(base, 0, reg->e_row, reg->e_col + 1, neg);
  }
  if(!neg) {
    delete base->clip_color;
    base->clip_color = 0;
  }
}

void HandleFocusChange(Widget w, XtPointer, XEvent *event, Boolean *)
{
  basewindow *base;
  
  for(base = basewindow::get_bases(); base; base = base->get_next())
    if(base->widget == w)
      break;
  
  if(event->type == FocusIn && base) {
    if(window::get_active_win()->get_base() != base)
      window::get_active_win()->next_window_active(base->get_windows(), FALSE);
    XMapWindow(MainDisplay, XtWindow(w));
  }
}

void HandleKeyPress(Widget, XtPointer, XEvent *o_event, Boolean *)
{
  XKeyEvent *event = (XKeyEvent *)o_event;
  
  if(event->type == KeyPress) {
    ReadUser(event);
  }
}

Boolean convertSelectionProc(Widget /* w */, Atom * /* XA_PRIMARY */,
			     Atom *target,
			     Atom *type_return, XtPointer *value_return,
			     unsigned long *length_return, int *format_return)
{
  if(*target != XA_STRING || !selection)
    return(FALSE);
  
  *type_return   = XA_STRING;
  *value_return  = selection;
  *length_return = selection_len;
  *format_return = 8;
  
  return(TRUE);
}

void loseSelectionProc(Widget /* w */, Atom * /* XA_PRIMARY selection */)
{
  if(selection) {
    delete[] selection;
    selection = 0;
  }
  selection_len = 0;
}

void doneSelectionProc(Widget /* widget */,
		       Atom * /* selection */, Atom * /* target */)
{
  /* this is needed so that area isn't freed after wards */
}

void clip_to_xbuf(basewindow *base, const int x, const int y)
{
  char *str;
  uchar letter;
  int s_col, e_col, s_row, e_row, amount, i, j, k, pos;
  
  s_col = x/base->font->max_bounds.width;
  if(s_col >= base->cols)
    s_col = base->cols - 1;
  if(s_col < 0)
    s_col = 0;
  
  s_row = y/(base->font->ascent + base->font->descent);
  if(s_row >= base->rows)
    s_row = base->rows - 1;
  if(s_row < 0)
    s_row = 0;
  
  if(s_row == base->h_row) {
    e_row = base->h_row;
    if(s_col > base->h_col) {
      e_col = s_col;
      s_col = base->h_col;
    } else {
      e_col = base->h_col;
    }
  } else if(s_row > base->h_row) {
    e_row = s_row;
    e_col = s_col;
    s_row = base->h_row;
    s_col = base->h_col;
  } else {
    e_row = base->h_row;
    e_col = base->h_col;
  }
  
  if(style == 1) {
    if(ownisalnum(base->data[s_row][s_col].letter)) {
      while(s_col > 0 && ownisalnum(base->data[s_row][s_col - 1].letter))
	s_col--;
      
    } else if(isspace(base->data[s_row][s_col].letter)) {
      while(s_col > 0 && isspace(base->data[s_row][s_col - 1].letter))
	s_col--;
      
    } else {
      letter = base->data[s_row][s_col].letter;
      while(s_col > 0 && base->data[s_row][s_col - 1].letter == letter)
	s_col--;
    }
    
    if(ownisalnum(base->data[e_row][e_col].letter)) {
      while(e_col < base->cols - 1 &&
	    ownisalnum(base->data[e_row][e_col + 1].letter))
	e_col++;
      
    } else if(isspace(base->data[e_row][e_col].letter)) {
      while(e_col < base->cols - 1 &&
	    isspace(base->data[e_row][e_col + 1].letter))
	e_col++;

    } else {
      letter = base->data[e_row][e_col].letter;
      while(e_col < base->cols - 1 &&
	    base->data[e_row][e_col + 1].letter == letter)
	e_col++;
    }
    
  } else if(style == 2) {
    s_col = 0;
    e_col = base->cols - 1;
  }
  
  if(base->clip_color) {
    if(base->clip_color->s_col == s_col && base->clip_color->e_col == e_col &&
       base->clip_color->s_row == s_row && base->clip_color->e_row == e_row)
      return;

    if(base->clip_color->s_row != s_row || base->clip_color->s_col != s_col) {
      if(base->clip_color->s_row == s_row) {
	if(base->clip_color->s_col < s_col) {
	  xprint_many(base, base->clip_color->s_col, s_row,
		      s_col - base->clip_color->s_col, FALSE);
	} else {
	  xprint_many(base, s_col, s_row, base->clip_color->s_col - s_col, TRUE);
	}
      } else {
	if(base->clip_color->s_row < s_row) {
	  xprint_many(base, base->clip_color->s_col, base->clip_color->s_row,
		      base->cols - base->clip_color->s_col, FALSE);
	  for(i = base->clip_color->s_row + 1; i < s_row; i++)
	    xprint_many(base, 0, i, base->cols, FALSE);
	  xprint_many(base, 0, s_row, s_col, FALSE);
	} else {
	  xprint_many(base, s_col, s_row, base->cols - s_col, TRUE);
	  for(i = s_row + 1; i < base->clip_color->s_row; i++)
	    xprint_many(base, 0, i, base->cols, TRUE);
	  xprint_many(base, 0, base->clip_color->s_row, base->clip_color->s_col,
		      TRUE);
	}
      }
    }
    
    if(base->clip_color->e_row != e_row || base->clip_color->e_col != e_col) {
      if(base->clip_color->e_row == e_row) {
	if(base->clip_color->e_col < e_col) {
	  
	  xprint_many(base, base->clip_color->e_col + 1, e_row,
		      e_col - base->clip_color->e_col, TRUE);
	} else {
	  xprint_many(base, e_col + 1, e_row, base->clip_color->e_col - e_col,
		      FALSE);
	}
      } else {
	if(base->clip_color->e_row < e_row) {
	  xprint_many(base, base->clip_color->e_col + 1, base->clip_color->e_row,
		      base->cols - base->clip_color->e_col - 1, TRUE);
	  for(i = base->clip_color->e_row + 1; i < e_row; i++)
	    xprint_many(base, 0, i, base->cols, TRUE);
	  xprint_many(base, 0, e_row, e_col + 1, TRUE);
	} else {
	  xprint_many(base, e_col + 1, e_row, base->cols - e_col - 1, FALSE);
	  for(i = e_row + 1; i < base->clip_color->e_row; i++)
	    xprint_many(base, 0, i, base->cols, FALSE);
	  
	  xprint_many(base, 0, base->clip_color->e_row,
		      base->clip_color->e_col + 1, FALSE);
	}
      }
    }
    base->clip_color->s_col = s_col;
    base->clip_color->e_col = e_col;
    base->clip_color->s_row = s_row;
    base->clip_color->e_row = e_row;
  } else {
    base->clip_color = new color_clip_reg;
    base->clip_color->s_col = s_col;
    base->clip_color->e_col = e_col;
    base->clip_color->s_row = s_row;
    base->clip_color->e_row = e_row;
    
    xprint_reg(base, TRUE);
  }
  
  amount = 0;
  if(s_row == e_row) {
    amount = e_col - s_col + 1;
  } else {
    amount = base->cols - s_col + 1;
    amount += e_col + 1;
    amount += (base->cols + 1) * (e_row - s_row - 1);
  }
  if(!(str = new char[amount])) {
    master_table.main_session->print((const uchar *)"#Out of memory\n\r");
    return;
  }
  if(s_row == e_row) {
    for(i = s_col, pos = 0; i <= e_col; i++, pos++)
      *(str + pos) = (char)base->data[s_row][i].letter;
  } else {
    pos = 0;
    k = base->cols;
    while(k > 0 && base->data[s_row][k - 1].letter == ' ')
      k--;
    
    for(i = s_col; i < k; i++, pos++)
      *(str + pos) = (char)base->data[s_row][i].letter;
    *(str + pos++) = '\n';
    
    for(i = s_row + 1; i < e_row; i++) {
      k = base->cols;
      while(k > 0 && base->data[i][k - 1].letter == ' ')
	k--;
      for(j = 0; j < k; j++, pos++)
	*(str + pos) = (char)base->data[i][j].letter;
      *(str + pos++) = '\n';
    }
    k = e_col + 1;
    while(k > 0 && base->data[s_row][k - 1].letter == ' ')
      k--;
    
    for(i = 0; i < k; i++, pos++)
      *(str + pos) = (char)base->data[e_row][i].letter;
  }
  
  if(selection)
    free(selection);
  selection = (uchar *)str;
  selection_len = pos;
  
  if(!XtOwnSelection(base->widget, XA_PRIMARY, CurrentTime,
		     convertSelectionProc, loseSelectionProc,
		     doneSelectionProc)) {
    delete[] str;
    selection = 0;
    selection_len = 0;
  }
}

void HandleMousePress(Widget w, XtPointer, XEvent *o_event, Boolean *)
{
  int prompt_len, font_width, font_height;
  long time_dif;
  window *activ;
  XButtonEvent *event = (XButtonEvent *)o_event;
  basewindow *base;
  
  struct timeval now;
  static struct timeval last_time = {0, 0};
  
  for(base = basewindow::get_bases(); base; base = base->get_next())
    if(base->widget == w)
      break;
  
  if(event->type == ButtonPress && base) {
    font_width  = base->font->max_bounds.width;
    font_height = base->font->ascent + base->font->descent;
    
    gettimeofday(&now, 0);
    
    time_dif  = now.tv_sec  - last_time.tv_sec;
    if(time_dif > 10)
      time_dif = 10;
    time_dif *= 1000000;
    time_dif += now.tv_usec - last_time.tv_usec;
    time_dif /= 1000; // milliseconds
    
    last_time.tv_sec  = now.tv_sec;
    last_time.tv_usec = now.tv_usec;
    
    if(event->button == Button1) {
      if(button == 1 && time_dif <= STYLE_MAX_TIME) {
	style++;
	if(style > 2)
	  style = 0;
      } else {
	button = 1;
	style = 0;
      }
      
      if(base->clip_color) {
	xprint_reg(base, FALSE);
      }
      base->h_col = event->x/font_width;
      if(base->h_col >= base->cols)
	base->h_col = base->cols - 1;
      base->h_row = event->y/font_height;
      if(base->h_row >= base->rows)
	base->h_row = base->rows - 1;
      
      if(style > 0) {
	clip_to_xbuf(base, event->x, event->y);
      }
 
    } else if(event->button == Button2) {
      activ = window::get_active_win();
      prompt_len = activ->get_active_ses()->get_prompt_len();
      
      func_insert_clip(prompt_len, activ);

    } else if(event->button == Button3) {
      if(button == 3 && time_dif <= STYLE_MAX_TIME) {
	style++;
	if(style > 2)
	  style = 0;
      } else {
	button = 3;
	style = 0;
      }
      clip_to_xbuf(base, event->x, event->y);
    }
  }
}

void HandleMouseMove(Widget w, XtPointer, XEvent *o_event, Boolean *)
{
  XMotionEvent *event = (XMotionEvent *)o_event;
  basewindow *base;
  
  for(base = basewindow::get_bases(); base; base = base->get_next())
    if(base->widget == w)
      break;
  
  if(event->type == MotionNotify && base) {
    clip_to_xbuf(base, event->x, event->y);
  }
}

void GetColors()
{
  char errmsg[256];
  int i;
  XColor exact, color;
  
  if(MainDisplay) {
    if(!XAllocNamedColor(MainDisplay,
			 DefaultColormap(MainDisplay,
					 DefaultScreen(MainDisplay)),
			 color_names[BLACK], &exact, &color)) {
      fprintf(stderr, "ERROR Couldn't get %s color.\n", color_names[BLACK]);
      exit(1);
    }
    colors[BLACK] = color.pixel;
    
    if(!XAllocNamedColor(MainDisplay,
			 DefaultColormap(MainDisplay,
					 DefaultScreen(MainDisplay)),
			 color_names[WHITE], &exact, &color)) {
      fprintf(stderr, "ERROR Couldn't get %s color.\n", color_names[WHITE]);
      exit(1);
    }
    colors[WHITE] = color.pixel;
    
    for(i = 1; i < 16; i++) {
      if(i != WHITE && i != BLACK) {
	if(!XAllocNamedColor(MainDisplay,
			     DefaultColormap(MainDisplay,
					     DefaultScreen(MainDisplay)),
			     color_names[i], &exact, &color)) {
	  sprintf(errmsg, "ERROR Couldn't get %s color.", color_names[i]);
	  filelog(errmsg);
	  colors[i] = colors[WHITE];
	  /* use white color all those witch couln't get */
	} else
	  colors[i] = color.pixel;
      }
    }
  } else {
    for(i = 0; i < 16; i++)
      colors[i] = 0;
  }
}

void FreeColors()
{
  int i;
  
  if(MainDisplay) {
    for(i = 0; i < 16; i++)
      if(colors[i] != colors[WHITE])
	XFreeColors(MainDisplay,
		    DefaultColormap(MainDisplay, DefaultScreen(MainDisplay)),
		    &colors[i], 1, (unsigned long)0);
    XFreeColors(MainDisplay,
		DefaultColormap(MainDisplay, DefaultScreen(MainDisplay)),
		&colors[WHITE], 1, (unsigned long)0);
  }
}

XFontStruct *GetFont(const char *fontname, session *ses)
{
  uchar buffer[BUFFER_SIZE];
  XFontStruct *tmp;
  
  if(fontname && *fontname) {
    if(!(tmp = XLoadQueryFont(MainDisplay, fontname))) {
      sprintf((char *)buffer, "##Couldn't load font '%s'\n\r", fontname);
      if(ses)
	ses->print(buffer);
      else
	filelog((const char *)buffer);
    } else {
      if(!def_font)
	def_font = tmp;
      
      return(tmp);
    }
  }
  
  if(!def_font) {
    if(!(def_font = XLoadQueryFont(MainDisplay, "fixed"))) {
      filelog("##Couldn't load font 'fixed'");
      exit_tintin(1);
    }
  }
  return(def_font);
}

void FreeFont(XFontStruct *f)
{
  if(f) {
    if(f != def_font)
      XFreeFont(MainDisplay, f);
  } else if(def_font) {
    XFreeFont(MainDisplay, def_font);
    def_font = 0;
  }
}
